@@ -297,10 +297,12 @@ postgres_get_resultset(const char *zone, const char *record,
 
 	REQUIRE(*rs == NULL);
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d Getting DBI", dlz_thread_num);
+#endif
 
 	/* get db instance / connection */
 #ifdef ISC_PLATFORM_USETHREADS
@@ -318,10 +320,12 @@ postgres_get_resultset(const char *zone, const char *record,
 
 #endif /* ISC_PLATFORM_USETHREADS */
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d Got DBI - checking query", dlz_thread_num);
+#endif
 
 	/* if DBI is null, can't do anything else */
 	if (dbi == NULL) {
@@ -391,10 +395,12 @@ postgres_get_resultset(const char *zone, const char *record,
 		goto cleanup;
 	}
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d checked query", dlz_thread_num);
+#endif
 
 	/*
 	 * was a zone string passed?  If so, make it safe for use in
@@ -410,10 +416,12 @@ postgres_get_resultset(const char *zone, const char *record,
 		dbi->zone = NULL;
 	}
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d did zone", dlz_thread_num);
+#endif
 
 	/*
 	 * was a record string passed?  If so, make it safe for use in
@@ -430,10 +438,12 @@ postgres_get_resultset(const char *zone, const char *record,
 	}
 
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d did record", dlz_thread_num);
+#endif
 
 	/*
 	 * was a client string passed?  If so, make it safe for use in
@@ -449,10 +459,12 @@ postgres_get_resultset(const char *zone, const char *record,
 		dbi->client = NULL;
 	}
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 	DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d did client", dlz_thread_num);
+#endif
 
 	/*
 	 * what type of query are we going to run?
@@ -486,10 +498,12 @@ postgres_get_resultset(const char *zone, const char *record,
 		goto cleanup;
 	}
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d built query", dlz_thread_num);
+#endif
 
 	/* if the querystring is null, Bummer, outta RAM.  UPGRADE TIME!!!   */
 	if (querystring  == NULL) {
@@ -497,10 +511,12 @@ postgres_get_resultset(const char *zone, const char *record,
 		goto cleanup;
 	}
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d query is '%s'", dlz_thread_num, querystring);
+#endif
 
 	/*
 	 * output the full query string during debug so we can see
@@ -512,11 +528,13 @@ postgres_get_resultset(const char *zone, const char *record,
 
 	/* attempt query up to 3 times. */
 	for (j=0; j < 3; j++) {
+#if 0
 		/* temporary logging message */
 		isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 			      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 			      "%d executing query for %d time",
 			      dlz_thread_num, j);
+#endif
 		/* try to get result set */
 		*rs = PQexec((PGconn *)dbi->dbconn, querystring );
 		result = ISC_R_SUCCESS;
@@ -525,11 +543,13 @@ postgres_get_resultset(const char *zone, const char *record,
 		 * attempts.
 		 */
 		for (i=0; *rs == NULL && i < 3; i++) {
+#if 0
 			/* temporary logging message */
 			isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 				      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 				      "%d resetting connection",
 				      dlz_thread_num);
+#endif
 			result = ISC_R_FAILURE;
 			PQreset((PGconn *) dbi->dbconn);
 			/* connection ok, break inner loop */
@@ -538,17 +558,21 @@ postgres_get_resultset(const char *zone, const char *record,
 		}
 		/* result set ok, break outter loop */
 		if (PQresultStatus(*rs) == PGRES_TUPLES_OK) {
+#if 0
 			/* temporary logging message */
 			isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 				      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 				      "%d rs ok", dlz_thread_num);
+#endif
 			break;
 		} else {
 			/* we got a result set object, but it's not right. */
+#if 0
 			/* temporary logging message */
 			isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 				      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 				      "%d clearing rs", dlz_thread_num);
+#endif
 			PQclear(*rs);	/* get rid of it */
 			/* in case this was the last attempt */
 			result = ISC_R_FAILURE;
@@ -558,10 +582,12 @@ postgres_get_resultset(const char *zone, const char *record,
  cleanup:
 	/* it's always good to cleanup after yourself */
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d cleaning up", dlz_thread_num);
+#endif
 
 	/* if we couldn't even allocate DBI, just return NULL */
 	if (dbi == NULL)
@@ -581,10 +607,12 @@ postgres_get_resultset(const char *zone, const char *record,
 
 #ifdef ISC_PLATFORM_USETHREADS
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d unlocking mutex", dlz_thread_num);
+#endif
 
 	/* release the lock so another thread can use this dbi */
 	isc_mutex_unlock(&dbi->instance_lock);
@@ -595,10 +623,12 @@ postgres_get_resultset(const char *zone, const char *record,
 	if (querystring  != NULL)
 		isc_mem_free(ns_g_mctx, querystring );
 
+#if 0
 	/* temporary logging message */
 	isc_log_write(dns_lctx, DNS_LOGCATEGORY_DATABASE,
 		      DNS_LOGMODULE_DLZ, ISC_LOG_ERROR,
 		      "%d returning", dlz_thread_num);
+#endif
 
 	/* return result */
 	return result;
@@ -986,12 +1016,15 @@ postgres_authority(const char *zone, void *driverarg, void *dbdata,
 /*% if zone is supported, lookup up a (or multiple) record(s) in it */
 static isc_result_t
 postgres_lookup(const char *zone, const char *name, void *driverarg,
-		void *dbdata, dns_sdlzlookup_t *lookup)
+		void *dbdata, dns_sdlzlookup_t *lookup,
+		dns_clientinfomethods_t *methods, dns_clientinfo_t *clientinfo)
 {
 	isc_result_t result;
 	PGresult *rs = NULL;
 
 	UNUSED(driverarg);
+	UNUSED(methods);
+	UNUSED(clientinfo);
 
 	/* run the query and get the result set from the database. */
 	result = postgres_get_resultset(zone, name, NULL, LOOKUP, dbdata, &rs);
@@ -1280,7 +1313,14 @@ static dns_sdlzmethods_t dlz_postgres_methods = {
 	postgres_lookup,
 	postgres_authority,
 	postgres_allnodes,
-	postgres_allowzonexfr
+	postgres_allowzonexfr,
+	NULL,
+	NULL,
+	NULL,
+	NULL,
+	NULL,
+	NULL,
+	NULL,
 };
 
 /*%
